// File management functions taken from the Fox library
// and modified to add a progress dialog 
// Some helper functions are also added to get large files support 

#include "config.h"
#include "i18n.h"

#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/wait.h>
#include <time.h>

// For Sun compatibility
#ifdef __sun
#include <alloca.h>
#endif

#include <fox-1.4/fx.h>
#include <fox-1.4/fxkeys.h>
#include <fox-1.4/FXPNGIcon.h>

#include "xfedefs.h"
#include "icons.h"
#include "File.h"
#include "OverwriteBox.h"
#include "MessageBox.h"
#include "CommandWindow.h"

#if defined(linux)
extern FXStringDict* mtdevices;
extern FXStringDict* updevices;
extern FXbool mount_messages;
#endif



// Delay before the progress bar should be shown (ms)
#define SHOW_PBAR_DELAY 1000

// Progress dialog width
#define PDIALOG_WIDTH 150

// Global variables
extern FXMainWindow *mainWindow;
extern FXbool confirm_overwrite;

// Timeout for stat functions (seconds)
#define TIMEOUT 5

// Safe strcpy function (Public domain, by C.B. Falconer)
// The destination string is always null terminated
size_t strlcpy(char *dst, const char *src, size_t sz)
{
	const char *start = src;

	if (src && sz--)
	{
		while ((*dst++ = *src))
			if (sz--)
				src++;
			else
		 	{
            	*(--dst) = '\0';
            	break;
         	}
   	}
   	if (src)
   	{
      	while (*src++)
			continue;
      	return src - start - 1;
   	}
   	else if (sz)
		*dst = '\0';
	return 0;
}


// Safe strcat function (Public domain, by C.B. Falconer)
// The destination string is always null terminated
size_t strlcat(char *dst, const char *src, size_t sz)
{
   char  *start = dst;

   while (*dst++)      // assumes sz >= strlen(dst)
      if (sz)
	  	sz--;          // i.e. well formed string
   dst--;
   return dst - start + strlcpy(dst, src, sz);
}


// To test if two strings are equal (strcmp replacement, thanks to Francesco Abbate)
int streq(const FXchar *a, const FXchar *b)
{
	if (a == NULL || b == NULL)
		return 0;
	return (strcmp(a, b) == 0);
}



// Replacement of the stat function with a timeout
int statout(const char *filename, struct stat *buf)
{
#if defined(linux)
	// It's a mount point
	if(mtdevices->find(filename))
	{
		// Mount point is up and is NFS or SMB
		if(::streq(updevices->find(filename),"up")
		   && (::streq(mtdevices->find(filename),"nfs") || ::streq(mtdevices->find(filename),"smbfs")))
		{
			int pid, ret;
			time_t t0, t;
	
			// Create child process
			pid = fork();
    		if (pid == -1)
        		return -1;
    
			if (pid == 0)
    		{
        		// Here, we are running as the child process!
				ret=stat(filename,buf);
				exit(ret);
    		}
    		else
    		{
        		// Here, we are running as the parent process!
    			time(&t0);
				while (1)
				{
					if ( (waitpid(pid,&ret,WNOHANG)==0 ) )
					{
						// Child is still running, wait
						time(&t);
				
						// Timeout exceeded, abort
						if (difftime(t,t0)>TIMEOUT)
						{
							kill(pid,SIGKILL);
							return -1;
						}
					}
					else
					{
						// Child has finished
						if (ret==0)
						{	
							ret=stat(filename,buf);
							return ret;
						}
						else
							return -1;
					}
				}
			}
		}
		// Mount point is down
		else if(::streq(updevices->find(filename),"down"))
			return -1;
		
		// Other mount point
		else
			return stat(filename,buf);		
	}
		
	// It's not a mount point
	else
#endif
	return stat(filename,buf);
}

// Replacement of the lstat function with a timeout
int lstatout(const char *filename, struct stat *buf)
{
#if defined(linux)
	// It's a mount point
	if(mtdevices->find(filename))
	{
		// Mount point is NFS or SMB
		if(::streq(updevices->find(filename),"up")
		   && (::streq(mtdevices->find(filename),"nfs") || ::streq(mtdevices->find(filename),"smbfs")))
		{
			int pid, ret;
			time_t t0, t;
	
			// Create child process
			pid = fork();
    		if (pid == -1)
       			return -1;
    
			if (pid == 0)
    		{
        		// Here, we are running as the child process!
				ret=lstat(filename,buf);
				exit(ret);
    		}
    		else
    		{
        		// Here, we are running as the parent process!
				time(&t0);
				while (1)
				{
					if ( (waitpid(pid,&ret,WNOHANG)==0 ) )
					{
						// Child is still running, wait
						time(&t);
				
						// Timeout exceeded, abort
						if (difftime(t,t0)>TIMEOUT)
						{
							kill(pid,SIGKILL);
							return -1;
						}
					}
					else
					{
						// Child has finished
						if (ret==0)
						{	
							ret=lstat(filename,buf);
							return ret;
						}
						else
							return -1;
					}
				}
			}
		}
		// Mount point is down
		else if(::streq(updevices->find(filename),"down"))
			return -1;
		
		// Other mount point
		else
			return lstat(filename,buf);		
	}
		
	// It's not a mount point
	else
#endif
	return lstat(filename,buf);
}


#if defined(linux)
// Stat function used to test if a mount point is up or down
int mtlstat(const char *filename, struct stat *buf)
{
	int pid, ret;
	time_t t0, t;
	
	// Create child process
	pid = fork();
    if (pid == -1)
       	return -1;
    
	if (pid == 0)
    {
        // Here, we are running as the child process!
		ret=lstat(filename,buf);
		exit(ret);
    }
    else
    {
        // Here, we are running as the parent process!
		time(&t0);
		while (1)
		{
			if ( (waitpid(pid,&ret,WNOHANG)==0 ) )
			{
				// Child is still running, wait
				time(&t);
				
				// Timeout exceeded, abort
				if (difftime(t,t0)>TIMEOUT)
				{
					kill(pid,SIGKILL);
					return -1;
				}
			}
			else
			{
				// Child has finished
				if (ret==0)
				{	
					ret=lstat(filename,buf);
					return ret;
				}
				else
					return -1;
			}
		}
	}
}
#endif


// Obtain the non recursive size of a directory
unsigned long long dirsize(char *path)
{
    DIR *dp;
	struct dirent *dirp;
	struct stat statbuf;
	char buf[MAXPATHLEN];
	unsigned long long dsize=0;
	int l, ret;
			
	if((dp=opendir(path))==NULL)
        return 0;
	
	l=strlen(path);
	if (l>0 && path[l-1]=='/')
		path[l-1]='\0';

	while((dirp=readdir(dp)))
    {
        if ( ::streq(dirp->d_name, ".")||::streq(dirp->d_name, "..") )
            continue;
		
		snprintf(buf,sizeof(buf)-1,"%s/%s",path,dirp->d_name);
		
#if defined(linux)
		if (mtdevices->find(buf))
			ret=lstatout(buf,&statbuf);
		else
#endif
			ret=lstat(buf,&statbuf);
		if(ret>=0)
		{
			if(!S_ISDIR(statbuf.st_mode))
		  		dsize += (unsigned long long) statbuf.st_size;
		}
	}
	if(closedir(dp)<0)
        fprintf(stderr,"Cannot close directory %s\n", path);
	return dsize;
}

// Obtain the recursive size of a directory
unsigned long long dirpath( char *path)
{
	struct stat statbuf;
	struct dirent *dirp;
    char *ptr;
    DIR *dp;
	unsigned long long dsize;
	int ret;

#if defined(linux)
	if (mtdevices->find(path))
		ret=lstatout(path,&statbuf);
	else
#endif
		ret=lstat(path,&statbuf);
	if(ret<0)
        return 0;
    dsize = (unsigned long long) statbuf.st_size;
    
	// Not a directory
	if(!S_ISDIR(statbuf.st_mode))
        return dsize;
    
	ptr=path + strlen(path);
    if(ptr[-1]!='/')
    {
        *ptr++='/';
        *ptr='\0';
    }
    
	if((dp=opendir(path))==NULL)
        return 0;
    
	while((dirp=readdir(dp)))
    {
        if( ::streq(dirp->d_name, ".") || ::streq(dirp->d_name, "..") )
            continue;
        strlcpy(ptr, dirp->d_name, strlen(dirp->d_name)+1);
		
		// Recursive call
        dsize+=dirpath(path);
    }
	
    ptr[-1]='\0'; // ??
	
    if(closedir(dp)<0)
        fprintf(stderr,"Cannot close directory %s\n", path);
    return dsize;
}

// Write the file size in human readable form (bytes or Kbytes or Mbytes)
FXString hSize(FXchar* size)
{
    int flag=0;
    FXchar suf[64];
	FXchar buf[128];
	FXString hsize;

    unsigned long long lsize = strtoull(size,NULL,10);
    float fsize=0.0;

	strlcpy(suf,_("bytes"),sizeof(suf)); 
	if(lsize>1000000000)
    {
        fsize=lsize/1073741824.0;
        strlcpy(suf,_("GB"),sizeof(suf));
        flag=1;
    }
	else if(lsize>1000000)
    {
        fsize=lsize/1048576.0;
        strlcpy(suf,_("MB"),sizeof(suf));
        flag=1;
    }
    else if(lsize>1000)
    {
        fsize=lsize/1024.0;
        strlcpy(suf,_("KB"),sizeof(suf));
        flag=1;
    }
  	
	if(flag)
    {
		if(fsize==(int)fsize)
            snprintf(buf,sizeof(buf),"%.0f %s",fsize,suf);
        else
            snprintf(buf,sizeof(buf),"%.1f %s",fsize,suf);
    }
    else
        snprintf(buf,sizeof(buf),"%llu %s",lsize,suf);
	
	hsize=buf;
	return hsize;
}


// Remove terminating '/' on a path string to simplify a file or directory path
// Thus '/bla/bla////' becomes '/bla/bla'
// Special case : '/' stays to '/'
// And then return the absolute path, based on the current directory path
FXString filePath(const FXString path)
{
	FXString in=path, out=path;
	while (1)
	{
		if (in[in.length()-1]=='/' & in.length()!=1)
		{
			out=in.trunc(in.length()-1);
			in=out;
		}
		else
			break;
	}
	FXString dir=FXFile::getCurrentDirectory();

	// If absolute path
	if(ISPATHSEP(out[0]))
		return (out);
	else
		return (dir+PATHSEPSTRING+out);
}

// Remove terminating '/' on a path string to simplify a file or directory path
// Thus '/bla/bla////' becomes '/bla/bla'
// Special case : '/' stays to '/'
// And then return the absolute path, based on the specified directory path
FXString filePath(const FXString path, const FXString dir)
{
	FXString in=path, out=path;
	while (1)
	{
		if (in[in.length()-1]=='/' && in.length()!=1)
		{
			out=in.trunc(in.length()-1);
			in=out;
		}
		else
			break;
	}
	// If absolute path
	if(ISPATHSEP(out[0]))
		return (out);
	else
		return (dir+PATHSEPSTRING+out);
}


// Obtain a unique file name based on the current time stamp, such as : filename__YYYYMMDD-HH:MM:SS-sec
// If filename is a path name, only the modified filename is returned
// Used for file deletion to trash can
FXString stampname(const FXString filename)
{
	struct timeval tv;
	char stamp[32];
	
	gettimeofday(&tv,NULL);
	snprintf(stamp,sizeof(stamp)-1,"__%s-%ld",FXFile::time("%Y%m%d-%H:%M:%S",tv.tv_sec).text(),tv.tv_usec);
	return (FXFile::name(filename)+stamp);
}


// Enquote filename to make it safe for shell
// (the function provided by Fox seems to be buggy??)
FXString quote(const FXString& file,FXbool forcequotes)
{
	FXString result;
	register FXint i,c;
	for(i=0; (c=file[i])!='\0'; i++)
	{
    	switch(c)
		{
      	case '"':              // Quote needs to be escaped
        	result+="\\\"";
        break;
      	case '\\':              // Backspace needs to be escaped, of course
        	result+="\\\\";
        break;
      	case '#':
      	case '~':
        	if(i) goto noquote;   // Only quote if at begin of filename
      	case '!':               // Special in csh
      	case '\'':
      	case '$':               // Variable substitution
      	case '&':
      	case '(':
      	case ')':
      	case ';':
      	case '<':               // Redirections, pipe
      	case '>':
      	case '|':
      	case '`':               // Command substitution
      	case '^':               // Special in sh
      	case '*':               // Wildcard characters
      	case '?':
      	case '[':
      	case ']':
      	case '\t':              // White space
      	case '\n':
      	case ' ':
        	forcequotes=TRUE;
      	default:                // Normal characters just added
			noquote:result+=c;
        break;
		}
	}
	if(forcequotes) return "\""+result+"\"";
	return result;
}

// Decode filename to get original again
FXString dequote(const FXString& file)
{
	FXString result;
	register FXint i,c;
	i=0;
	while((c=file[i])!='\0' && isspace((FXuchar)c)) i++;
  	if(file[i]=='\"')
	{
    	i++;
    	while((c=file[i])!='\0' && c!='\"')
		{
      		if(c=='\\' && file[i+1]!='\0') c=file[++i];
      		result+=c;
      		i++;
      	}
    }
	else
	{
    	while((c=file[i])!='\0' /*&& !isspace((FXuchar)c)*/)
		{
      		if(c=='\\' && file[i+1]!='\0') c=file[++i];
      		result+=c;
      		i++;
      	}
    }
  	return result;
}

// Test if a directory is empty
// Return -1 if not a directory, 1 if empty and 0 if not empty
FXint isEmptyDir(const FXString directory)
{
	FXint ret=-1;
	DIR* dir;
	struct dirent *entry;
	int n=0;
	
	if ((dir=opendir(directory.text()))!=NULL)
	{
		while (n<3)	{entry=readdir(dir); n++;}
		if (entry==NULL)
			ret=1;
		else
			ret=0;
	}
	closedir(dir);
	return ret;
}
		
// Check if file exists
FXbool exists(const FXString& file)
{
  	struct stat status;
  	return !file.empty() && (lstatout(file.text(),&status)==0);
}

// Check if the file represents a directory
FXbool isDirectory(const FXString& file)
{
  	struct stat status;
  	return !file.empty() && (statout(file.text(),&status)==0) && S_ISDIR(status.st_mode);
}

// Check if file represents a file
FXbool isFile(const FXString& file)
{
  	struct stat status;
  	return !file.empty() && (statout(file.text(),&status)==0) && S_ISREG(status.st_mode);
}

// Check if the file or the link refered file is writable
FXbool isWritable(const FXString& file)
{
  	struct stat status;
	FXuint uid=getuid();
	FXuint gid=getgid();
	
	// File exists and can be stated
	if (!file.empty() && (statout(file.text(),&status)==0))
	{
		// We are root
		if (uid==0)
		{
			return TRUE;
		}
		// We are owner of the file
		else if (uid==status.st_uid)
		{
			// The file is writable
			if (status.st_mode & S_IWUSR)
				return TRUE;
			else
				return FALSE;
		}
		// We are in the group of the file
		else if (gid==status.st_gid)
		{
			// The file is writable
			if (status.st_mode & S_IWGRP)
				return TRUE;
			else
				return FALSE;
		}
		// We are in others
		else
		{
			// The file is writable
			if (status.st_mode & S_IWOTH)
				return TRUE;
			else
				return FALSE;
		}
	}	
	
	// File doesn't exist or cannot be stated
	else
		return FALSE;
}

// Check if the file or the link itself is writable
FXbool isLWritable(const FXString& file)
{
  	struct stat status;
	FXuint uid=getuid();
	FXuint gid=getgid();
	
	// File or link exists and can be stated
	if (!file.empty() && (lstatout(file.text(),&status)==0))
	{
		// We are root
		if (uid==0)
		{	
			return TRUE;
		}
		// We are owner of the file or link
		else if (uid==status.st_uid)
		{
			// The file or link is writable
			if (status.st_mode & S_IWUSR)
				return TRUE;
			else
				return FALSE;
		}
		// We are in the group of the file or link
		else if (gid==status.st_gid)
		{
			// The file or link is writable
			if (status.st_mode & S_IWGRP)
				return TRUE;
			else
				return FALSE;
		}
		// We are other
		else
		{
			// The file or link is writable
			if (status.st_mode & S_IWOTH)
				return TRUE;
			else
				return FALSE;
		}
	}	
	
	// File or link doesn't exist or cannot be stated
	else
		return FALSE;
}

// Check if the file or the link refered file is readable AND executable
// Function used to test if we can enter a directory
FXbool isReadExecutable(const FXString& file)
{
  	struct stat status;
	FXuint uid=getuid();
	FXuint gid=getgid();
	
	// File exists and can be stated
	if (!file.empty() && (statout(file.text(),&status)==0))
	{
		// We are root
		if (uid==0)
		{
			return TRUE;
		}
		// We are owner of the file
		else if (uid==status.st_uid)
		{
			// The file is readable and executable
			if ((status.st_mode & S_IRUSR) && (status.st_mode & S_IXUSR))
				return TRUE;
			else
				return FALSE;
		}
		// We are in the group of the file
		else if (gid==status.st_gid)
		{
			// The file is readable and executable
			if ((status.st_mode & S_IRGRP) && (status.st_mode & S_IXGRP))
				return TRUE;
			else
				return FALSE;
		}
		// We are in others
		else
		{
			// The file is readable and executable
			if ((status.st_mode & S_IROTH) && (status.st_mode & S_IXOTH))
				return TRUE;
			else
				return FALSE;
		}
	}	
	
	// File doesn't exist or cannot be stated
	else
		return FALSE;
}

// Check if file represents a link
FXbool isLink(const FXString& file)
{
  	struct stat status;
  	return !file.empty() && (lstatout(file.text(),&status)==0) && S_ISLNK(status.st_mode);
}

// Get file info
FXbool info(const FXString& file,struct stat& inf)
{
  	return !file.empty() && (statout(file.text(),&inf)==0);
}

// Read symbolic link
FXString readlink(const FXString& file)
{
	FXchar lnk[MAXPATHLEN+1];
	FXint len=readlink(file.text(),lnk,MAXPATHLEN);
	if(0<=len)
  		return FXString(lnk,len);
  	else
  		return FXString::null;
}


// Return true if files are identical
// Compare file names and inodes for case insensitive filesystems
FXbool identical(const FXString& file1,const FXString& file2)
{
	if(file1!=file2)
	{
    	struct stat stat1,stat2;
    	return !::lstatout(file1.text(),&stat1) && !::lstatout(file2.text(),&stat2) && stat1.st_ino==stat2.st_ino && stat1.st_dev==stat2.st_dev;
    }
  	return TRUE;
}


// Message Map
FXDEFMAP(File) FileMap[]={
                             FXMAPFUNC(SEL_COMMAND,File::ID_CANCEL_BUTTON,File::onCmdCancel),
							 FXMAPFUNC(SEL_TIMEOUT,File::ID_TIMEOUT,File::onTimeout),
                         };

// Object implementation
FXIMPLEMENT(File,FXDialogBox,FileMap,ARRAYNUMBER(FileMap))

// Construct object
File::File(FXWindow *owner, FXString title, const unsigned int operation):FXDialogBox(owner,title,DECOR_TITLE|DECOR_BORDER)
{
    // Progress window
	FXPacker *buttons=new FXPacker(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X,0,0,10,10,PDIALOG_WIDTH,PDIALOG_WIDTH,5,5);
    new FXHorizontalSeparator(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|SEPARATOR_GROOVE);
    FXVerticalFrame *contents=new FXVerticalFrame(this,LAYOUT_SIDE_TOP|FRAME_NONE|LAYOUT_FILL_X|LAYOUT_FILL_Y);
	allowPdialog=TRUE;

    // Cancel Button
    cancelButton=new FXButton(buttons,_("&Cancel"),NULL,this,File::ID_CANCEL_BUTTON,FRAME_RAISED|FRAME_THICK|LAYOUT_CENTER_X,0,0,0,0,20,20);
    cancelButton->setFocus();
    cancelButton->addHotKey(KEY_Escape);
	isCancelled=FALSE;

	// Progress bar
	pbar=NULL;

	// Progress dialog depends on the file operation
	switch(operation)	
	{
		case COPY:
	    	// Labels and progress bar
			uplabel=new FXLabel(contents,_("Source :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
    		downlabel=new FXLabel(contents,_("Target :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
			pbar=new FXProgressBar(contents,NULL,0,LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK|PROGRESSBAR_PERCENTAGE,0,0,0,0,PDIALOG_WIDTH);
		
			// Timer on
			getApp()->addTimeout(this,File::ID_TIMEOUT,SHOW_PBAR_DELAY);
			break;

    	case MOVE:
	    	// Labels and progress bar
			uplabel=new FXLabel(contents,_("Source :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
    		downlabel=new FXLabel(contents,_("Target :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
			pbar=new FXProgressBar(contents,NULL,0,LAYOUT_FILL_X|FRAME_SUNKEN|FRAME_THICK|PROGRESSBAR_PERCENTAGE,0,0,0,0,PDIALOG_WIDTH);

			// Timer on
			getApp()->addTimeout(this,File::ID_TIMEOUT,SHOW_PBAR_DELAY);
			break;

    	case DELETE:
	    	// Labels
			uplabel=new FXLabel(contents,_("Delete :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
    		downlabel=new FXLabel(contents,_("From :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);

			// Timer on
			getApp()->addTimeout(this,File::ID_TIMEOUT,SHOW_PBAR_DELAY);
			break;

    	case CHMOD:
	    	// Labels
			uplabel=new FXLabel(contents,_("Changing permissions..."),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
    		downlabel=new FXLabel(contents,_("File :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);

			// Timer on
			getApp()->addTimeout(this,File::ID_TIMEOUT,SHOW_PBAR_DELAY);
			break;

    	case CHOWN:
	    	// Labels
			uplabel=new FXLabel(contents,_("Changing owner..."),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
    		downlabel=new FXLabel(contents,_("File :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);

			// Timer on
			getApp()->addTimeout(this,File::ID_TIMEOUT,SHOW_PBAR_DELAY);
			break;

#if defined(linux)
    	case MOUNT:
	    	// Labels
			uplabel=new FXLabel(contents,_("Mount file system..."),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
    		downlabel=new FXLabel(contents,_("Mount the folder :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
			break;

    	case UNMOUNT:
	    	// Labels
			uplabel=new FXLabel(contents,_("Unmount file system..."),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
    		downlabel=new FXLabel(contents,_("Unmount the folder :"),NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
			break;
#endif

		default: // Other : RENAME, SYMLINK, ARCHIVE, EXTRACT, RPM_INSTALL, RPM_UNINSTALL
			// Progress dialog not used
			uplabel=NULL;
    		downlabel=NULL;
	}
	
	// Initialize the overwrite flags
	if (confirm_overwrite)
	{
		overwrite=FALSE;
		overwrite_all=FALSE;
		skip_all=FALSE;
	}
	else
	{
		overwrite=TRUE;
		overwrite_all=TRUE;
		skip_all=FALSE;
	}		    			
}

// Destructor
File::~File()
{
	getApp()->removeTimeout(this,File::ID_TIMEOUT);
    delete pbar;
}



// Create and initialize
void File::create()
{
	FXDialogBox::create();
}

// Force check of timeout for progress dialog (to avoid latency problems)
int File::checkTimeout(void)
{
	timeval now;	
	if (getApp()->hasTimeout(this,File::ID_TIMEOUT))
	{
		if (getApp()->remainingTimeout(this,File::ID_TIMEOUT)==0)
			if (now.tv_sec>0)
			{
				getApp()->removeTimeout(this,File::ID_TIMEOUT);
				if (allowPdialog)
					show(PLACEMENT_OWNER);
				getApp()->forceRefresh();
				getApp()->flush();
				return 1;
			}
	}
	return 0;
}

// Read bytes
long File::fullread(int fd,unsigned char *ptr,long len)
{
    long nread;
#ifdef EINTR
    do
        nread=read(fd,ptr,len);
    while(nread<0 && errno==EINTR);
#else
    nread=read(fd,ptr,len);
#endif
    return nread;
}


// Write bytes
long File::fullwrite(int fd,const unsigned char *ptr,long len)
{
    long nwritten,ntotalwritten=0;
    while(len>0)
    {
        nwritten=write(fd,ptr,len);
        if(nwritten<0)
        {
#ifdef EINTR
            if(errno==EINTR)
                continue;
#endif
            return -1;
        }
        ntotalwritten+=nwritten;
        ptr+=nwritten;
        len-=nwritten;
    }
    return ntotalwritten;
}


// Copy ordinary file
FXbool File::copyfile(const FXString& source, const FXString& target)
{
    FXString destfile;
	unsigned char buffer[4096];
    struct stat status;
    long nread,nwritten;
	long size, nread_all=0;
	double pct=0;
    int src,dst;
    FXbool ok=FALSE;

    if((src=::open(source.text(),O_RDONLY))>=0)
    {
        if(statout(source.text(),&status)==0)
        {
			// If destination is a directory
			if(::isDirectory(target))
				destfile=target+PATHSEPSTRING+FXFile::name(source);
			else
				destfile=target;
				
         	// Copy file block by block
			size=status.st_size; 
			if((dst=::open(destfile.text(),O_WRONLY|O_CREAT|O_TRUNC,status.st_mode))>=0)
            {
                while(1)
                {
                    nread=File::fullread(src,buffer,sizeof(buffer));
                    if(nread<0)
					{
						MessageBox::error(this,MBOX_OK,_("Error"),"%s",strerror(errno));									
                        goto err;
					}
                    if(nread==0)
                        break;
											
					// Force timeout checking for progress dialog
					checkTimeout();

					// Set percentage value for progress dialog
					nread_all+=nread;
					pct=100*(double)nread_all/(double)size;	
					if (pbar)
						pbar->setProgress((int)pct);

					// Give cancel button an opportunity to be clicked
					if(cancelButton)
						getApp()->runModalWhileEvents(cancelButton);

					// Set labels for progress dialog
					FXString label=_("Source : ")+source;
					uplabel->setText(label);
					label=_("Target : ")+target;
					downlabel->setText(label);
					getApp()->repaint();

					// If cancel button was clicked, close files and return
					if (isCancelled)
					{
						::close(dst);
						::close(src);
						return FALSE;
					}
                    nwritten=File::fullwrite(dst,buffer,nread);
                    if(nwritten<0)
					{
						MessageBox::error(this,MBOX_OK,_("Error"),"%s",strerror(errno) );			
                        goto err;
					}
                }
                ok=TRUE;
err:
                ::close(dst);
            }
        }
        ::close(src);
    }
    return ok;
}


// Copy directory
FXbool File::copydir(const FXString& source,const FXString& target,struct stat& parentstatus,inodelist* inodes)
{
    DIR *dirp;
    struct dirent *dp;
    struct stat status;
    inodelist *in,inode;
    FXString destfile,oldchild,newchild;

	// If destination is an existing directory
	if(::isDirectory(target))
		destfile=target+PATHSEPSTRING+FXFile::name(source);
	else
		destfile=target;

    // See if visited this inode already
    for(in=inodes; in; in=in->next)
    {
        if(in->st_ino==parentstatus.st_ino)
            return TRUE;
    }

    // Try make directory, if none exists yet
    if(mkdir(destfile.text(),parentstatus.st_mode|S_IWUSR)!=0 && errno!=EEXIST)
        return FALSE;

    // Can we stat it
    if(lstatout(destfile.text(),&status)!=0 || !S_ISDIR(status.st_mode))
        return FALSE;

    // Try open directory to copy
    dirp=opendir(source.text());
    if(!dirp)
        return FALSE;

    // Add this to the list
    inode.st_ino=status.st_ino;
    inode.next=inodes;

    // Copy stuff
    while((dp=readdir(dirp))!=NULL)
    {
        if(dp->d_name[0]!='.' || (dp->d_name[1]!='\0' && (dp->d_name[1]!='.' || dp->d_name[2]!='\0')))
        {
            oldchild=source;
            if(!ISPATHSEP(oldchild[oldchild.length()-1]))
                oldchild.append(PATHSEP);
            oldchild.append(dp->d_name);
            newchild=destfile;
            if(!ISPATHSEP(newchild[newchild.length()-1]))
                newchild.append(PATHSEP);
            newchild.append(dp->d_name);
			if(!copyrec(oldchild,newchild,&inode))
            {
				// If the cancel button was pressed
				if (isCancelled)
				{
                	closedir(dirp);
                	return FALSE;
				}
				
				// Or a permission problem occured
				else
				{
					FXPNGIcon icon(getApp(),errorIcon);
					FXString str;
					if (::isDirectory(oldchild))
						str=_("Can't copy folder ") + oldchild + _(" : Permission denied");
					else
						str=_("Can't copy file ") + oldchild + _(" : Permission denied");
       				MessageBox box(this,_("Error"),str,&icon,MBOX_OK_CANCEL|DECOR_TITLE|DECOR_BORDER);        	
					FXuint answer=box.execute(PLACEMENT_SCREEN);
					if(answer == MBOX_CLICKED_CANCEL)
					{
               			closedir(dirp);
						isCancelled=TRUE;
                		return FALSE;
					}					
				}
            }
        }
    }

    // Close directory
    closedir(dirp);

    // Success
    return TRUE;
}


// Recursive copy
FXbool File::copyrec(const FXString& source,const FXString& target,inodelist* inodes)
{
    struct stat status1,status2;

    // Source file or directory does not exist
    if(lstatout(source.text(),&status1)!=0)
        return FALSE;

    // If target is not a directory, remove it if allowed
    if(lstatout(target.text(),&status2)==0)
    {
        if(!S_ISDIR(status2.st_mode))
        {
            if(!(overwrite|overwrite_all))
                return FALSE;
            if(unlink(target.text())!=0)
                return FALSE;
        }
    }

    // Source is directory: copy recursively
    if(S_ISDIR(status1.st_mode))
        return File::copydir(source,target,status1,inodes);

    // Source is regular file: copy block by block
    if(S_ISREG(status1.st_mode))
        return File::copyfile(source,target);

    // Source is fifo: make a new one
    if(S_ISFIFO(status1.st_mode))
        return mkfifo(target.text(),status1.st_mode);

    // Source is device: make a new one
    if(S_ISBLK(status1.st_mode) || S_ISCHR(status1.st_mode) || S_ISSOCK(status1.st_mode))
        return mknod(target.text(),status1.st_mode,status1.st_rdev)==0;

    // Source is symbolic link: make a new one
    if(S_ISLNK(status1.st_mode))
    {
        FXString lnkfile=readlink(source);
        return ::symlink(lnkfile.text(),target.text())==0;
    }

    // This shouldn't happen
    return FALSE;
}



// Copy file
FXbool File::copy(const FXString& source, const FXString& target, const FXbool confirm_dialog)
{
	FXString targetfile;

	// Source doesn't exist
	if(!::exists(source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s doesn't exist"), source.text());
		return FALSE;
	}

	// Source and target are identical
	if (::identical(target,source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s is identical to target"), target.text());
		return FALSE;
	}

	// Target is an existing directory
	if (::isDirectory(target))
		targetfile=target+PATHSEPSTRING+FXFile::name(source);
	else
		targetfile=target;

	// Source and target file are identical
	if (::identical(targetfile,source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s is identical to target"), targetfile.text());			
		return FALSE;
	}

	// Target already exists
	if(::exists(targetfile))
	{
		// Overwrite dialog if necessary
		if (!(overwrite_all | skip_all) & confirm_dialog)
    	{
			FXString label=_("Source : ")+source;
			uplabel->setText(label);
			label=_("Target : ")+targetfile;
			downlabel->setText(label);
			getApp()->repaint();

			FXString msg=_("File ")+targetfile+_(" already exists. Overwrite?");
       		ConfirmOverWriteBox* dlg = new ConfirmOverWriteBox(this,_("Confirm Overwrite"),msg);
       		FXuint answer=dlg->execute();
			delete dlg;
       		switch(answer)
       		{
       			case 1:
           			overwrite = TRUE;
           			break;
       			case 2:
           			overwrite_all = TRUE;
           			break;
       			case 3:
					overwrite = FALSE;
           			break;
       			case 4:
           			skip_all = TRUE;
					break;
       		}
		}
		if(!(overwrite | overwrite_all) | skip_all)
			return FALSE;

		// Use of the standard C function to remove the target
		// avoids to overwrite a non empty directory
		if (::remove(targetfile.text())==-1)
		{
			MessageBox::error(this,MBOX_OK,_("Error"),_("Cannot overwrite the non empty folder %s"), targetfile.text());			
			return FALSE;
		}
	}

	// Copy file or directory
	return File::copyrec(source,targetfile,NULL);
}


// Remove file or directory
FXbool File::remove(const FXString& file)
{
	FXString dirname;
    struct stat status;
	static FXbool ISDIR=FALSE;

    if(lstatout(file.text(),&status)==0)
    {
        // It is a directory
		if(S_ISDIR(status.st_mode))
        {
            DIR *dirp=opendir(file.text());
            if(dirp)
            {			
                struct dirent *dp;
                FXString child;
				
				// Used to display only one progress dialog when deleting a directory
				ISDIR=TRUE;
				
				// Force timeout checking for progress dialog
				checkTimeout();

				// Give cancel button an opportunity to be clicked
				if(cancelButton)
					getApp()->runModalWhileEvents(cancelButton);

				// Set labels for progress dialog
				FXString label=_("Delete folder :")+file;
				uplabel->setText(label);
				dirname=FXFile::directory(FXFile::absolute(file));
				label=_("From : ")+dirname;
				downlabel->setText(label);
				getApp()->repaint();

				// If cancel button was clicked, return
				if (isCancelled)
					return FALSE;

                while((dp=readdir(dirp))!=NULL)
                {
					if(dp->d_name[0]!='.' || (dp->d_name[1]!='\0' && (dp->d_name[1]!='.' || dp->d_name[2]!='\0')))
                    {
                        child=file;
                        if(!ISPATHSEP(child[child.length()-1]))
                            child.append(PATHSEP);
                        child.append(dp->d_name);
                        if(!File::remove(child))
						{
							closedir(dirp);
							return FALSE;
						}
                    }
                }
                closedir(dirp);
            }
			if (rmdir(file.text())==-1)
			{
				MessageBox::error(this,MBOX_OK,_("Error"),_("Cannot delete directory %s"), file.text());			
				return FALSE;
			}
			else
				return TRUE;
        }
        else
        {
			// If it was not a directory
			if (!ISDIR)
			{
				// Force timeout checking for progress dialog
				checkTimeout();

				// Give cancel button an opportunity to be clicked
				if(cancelButton)
					getApp()->runModalWhileEvents(cancelButton);

				// Set labels for progress dialog
				FXString label=_("Delete : ")+file;
				uplabel->setText(label);
				dirname=FXFile::directory(FXFile::absolute(file));
				label=_("From : ")+dirname;
				downlabel->setText(label);
				getApp()->repaint();

				// If cancel button was clicked, return
				if (isCancelled)
					return FALSE;
			}
			if (unlink(file.text())==-1)
			{
				MessageBox::error(this,MBOX_OK,_("Error"),_("Cannot delete file %s"), file.text());			
				return FALSE;
			}
			else
				return TRUE;
        }
    }
    return FALSE;
}


// Rename a file or a directory
FXbool File::rename(const FXString& source, const FXString& target)
{
	// Source doesn't exist
    if(!::exists(source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s doesn't exist"), source.text());
		return FALSE;
	}
			
	// Source and target are identical
	if (::identical(target,source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s is identical to target"), target.text());
		return FALSE;
	}

	// Target already exists
	if(::exists(target))
	{
		// Overwrite dialog if necessary
		if (!(overwrite_all | skip_all))
    	{
			FXString msg=_("File ")+target+_(" already exists. Overwrite?");
       		ConfirmOverWriteBox* dlg = new ConfirmOverWriteBox(this,_("Confirm Overwrite"),msg);
       		FXuint answer=dlg->execute();
			delete dlg;
       		switch(answer)
       		{
       			case 1:
           			overwrite = TRUE;
           			break;
       			case 2:
           			overwrite_all = TRUE;
           			break;
       			case 3:
					overwrite = FALSE;
           			break;
       			case 4:
           			skip_all = TRUE;
					break;
       		}
		}
		if(!(overwrite | overwrite_all) | skip_all)
			return FALSE;
	}
       		
	// Rename file using the standard C function
	// This should only work for files that are on the same file system
	if (::rename(source.text(),target.text())==0)
		return TRUE;
	if(errno!=EXDEV)
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Cannot rename to target %s"), target.text());	
		return FALSE;
	}

	// If files are on different file systems, use the copy/delete scheme
	FXbool ret=this->copy(source,target,FALSE);
	if (ret)
		return (remove(source.text())==TRUE);
	else
		return FALSE;
}


// Move files
FXbool File::move(const FXString& source,const FXString& target)
{
	// Source doesn't exist
	if(!::exists(source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s doesn't exist"), source.text());
		return FALSE;
	}

	// Source and target are identical
	if (identical(target,source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s is identical to target"), target.text());
		return FALSE;
	}

	// Target is an existing directory
	FXString targetfile;
	if (::isDirectory(target))
		targetfile=target+PATHSEPSTRING+FXFile::name(source);
	else
		targetfile=target;

	// Source and target file are identical
	if (::identical(targetfile,source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s is identical to target"), target.text());
		return FALSE;
	}

	// Force timeout checking for progress dialog
	checkTimeout();

	// Give cancel button an opportunity to be clicked
	if(cancelButton)
		getApp()->runModalWhileEvents(cancelButton);

	// Set labels for progress dialog
	FXString label=_("Source : ")+source;
	uplabel->setText(label);
	label=_("Target : ")+target;
	downlabel->setText(label);
	getApp()->repaint();

	// Target file already exists
	if(::exists(targetfile))
    {
		// Overwrite dialog if necessary
		if (!overwrite_all & !skip_all)
		{
			FXString msg=_("File ")+targetfile+_(" already exists. Overwrite?");
       		ConfirmOverWriteBox* dlg = new ConfirmOverWriteBox(this,_("Confirm Overwrite"),msg);
       		FXuint answer=dlg->execute();
			delete dlg;

       		switch(answer)
       		{
       			case 1:
           			overwrite = TRUE;
           			break;
       			case 2:
           			overwrite_all = TRUE;
           			break;
       			case 3:
           			overwrite = FALSE;
					break;
       			case 4:
           			skip_all = TRUE;
					break;
       		}
		}
		if(!(overwrite | overwrite_all) | skip_all)
			return FALSE;
	}

	// Set the progress to 100% for move file operation
	if (pbar)
		pbar->setProgress(100);

	// Rename file using the standard C function
	// This should only work for files that are on the same file system
	if (::rename(source.text(),targetfile.text())==0)
		return TRUE;
	if(errno!=EXDEV)
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Cannot rename to target %s"), targetfile.text());	
		return FALSE;
	}

	// If files are on different file systems, use the copy/delete scheme
	FXbool ret=this->copy(source,targetfile,FALSE);
	if (ret)
		return (remove(source.text())==TRUE);
	else
		return FALSE;
}


// Symbolic Link file
FXbool File::symlink(const FXString& source,const FXString& target)
{
	// Source doesn't exist
    if(!::exists(source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s doesn't exist"), source.text());
		return FALSE;
	}
			
	// Source and target are identical
	if (::identical(target,source))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Source %s is identical to target"), target.text());
		return FALSE;
	}

	// Target already exists
	if(::exists(target))
	{
		// Overwrite dialog if necessary
		if (!(overwrite_all | skip_all))
    	{
			FXString msg=_("File ")+target+_(" already exists. Overwrite?");
       		ConfirmOverWriteBox* dlg = new ConfirmOverWriteBox(this,_("Confirm Overwrite"),msg);
       		FXuint answer=dlg->execute();
			delete dlg;
       		switch(answer)
       		{
       			case 1:
           			overwrite = TRUE;
           			break;
       			case 2:
           			overwrite_all = TRUE;
           			break;
       			case 3:
					overwrite = FALSE;
           			break;
       			case 4:
           			skip_all = TRUE;
					break;
       		}
		}
		if(!(overwrite | overwrite_all) | skip_all)
			return FALSE;
	}
    
	// Create symbolic link using the standard C function
	FXint ret=::symlink(source.text(),target.text());	
	if (ret==0)
		return TRUE;
	else
	{
		MessageBox::error(this,MBOX_OK,_("Error"),"%s",strerror(errno));
		return FALSE;
	}		
}


// Chmod a file or directory, recursively or not
// Note : the variable file returns the last processed file
// It can be different from the initial path, if recursive chmod is used
// (Used to fill an error message, if needed) 
int File::chmod(char *path, char *file, mode_t mode, FXbool rec, const FXbool dironly)
{		
    struct stat stats;

	// Initialise the file variable with the initial path
	strlcpy(file, path, strlen(path)+1);

    if (lstatout(path, &stats)) //if it doesn't exist
        return -1;

    if (!S_ISDIR(stats.st_mode)) // file
	{
        if(dironly)
            return 0;
			
		// Force timeout checking for progress dialog
		checkTimeout();

		// Give cancel button an opportunity to be clicked
		if(cancelButton)
			getApp()->runModalWhileEvents(cancelButton);

		// Set labels for progress dialog
		FXString label=_("Changing permissions...");
		uplabel->setText(label);
		label=_("File : ")+FXString(path);
		downlabel->setText(label);
		getApp()->repaint();

		// If cancel button was clicked, return
		if (isCancelled)
			return -1;

        return ::chmod(path,mode);
    }
    else // directory
    {
        if(rec == FALSE)
        {
			// Force timeout checking for progress dialog
			checkTimeout();

			// Give cancel button an opportunity to be clicked
			if(cancelButton)
				getApp()->runModalWhileEvents(cancelButton);

			// Set labels for progress dialog
			FXString label=_("Changing permissions...");
			uplabel->setText(label);
			label=_("Folder : ")+FXString(path);
			downlabel->setText(label);
			getApp()->repaint();

			// If cancel button was clicked, return
			if (isCancelled)
				return -1;

			if(::chmod(path,mode)) // do not change recursively
				return -1;
        }
        else
            return rchmod(path,file,mode,dironly); // recursive change
    }
    return 0;

}


// Recursive chmod for a directory
int File::rchmod(char *path, char *file, mode_t mode, const FXbool dironly)
{
    struct stat stats;

	// Initialise the file variable with the initial path
	strlcpy(file, path, strlen(path)+1);

    if (lstatout(path, &stats)) //if it doesn't exist
        return -1;

    if (!S_ISDIR(stats.st_mode)) // file
    {
        if(dironly)
            return 0;

		// Force timeout checking for progress dialog
		checkTimeout();

		// Give cancel button an opportunity to be clicked
		if(cancelButton)
			getApp()->runModalWhileEvents(cancelButton);

		// Set labels for progress dialog
		FXString label=_("Changing permissions...");
		uplabel->setText(label);
		label=_("File : ")+FXString(path);
		downlabel->setText(label);
		getApp()->repaint();

		// If cancel button was clicked, return
		if (isCancelled)
			return -1;

        return ::chmod(path,mode);
    }

    DIR *dir;
    struct dirent *entry;
    int i, pl = strlen(path);

    if (!(dir = opendir(path)))
        return -1;

    for(i = 0; (entry = readdir(dir)); i++)
	{	
        if (entry->d_name[0] != '.' || (entry->d_name[1] != '\0'
                                        && (entry->d_name[1] != '.' ||
                                            entry->d_name[2] != '\0')))
        {
            int pl1 = pl, l = strlen(entry->d_name);
            char *path1 = (char *)alloca(pl1+l+2);

            strlcpy(path1, path, strlen(path)+1);
            if (path1[pl1-1] != '/')
                path1[pl1++] = '/';
            strlcpy(path1+pl1, entry->d_name, strlen(entry->d_name)+1);

			// Modify the file variable with the new path
			strlcpy(file, path1, strlen(path1)+1);
            if (rchmod(path1,file,mode,dironly))
                return -1;
        }
	}

    if (closedir(dir))
        return -1;

    return ::chmod(path,mode);
}


// Chown a file or directory, recursively or not
// Note : the variable file returns the last processed file
// It can be different from the initial path, if recursive chmod is used
// (Used to fill an error message, if needed) 
int File::chown(char *path, char *file, uid_t uid, gid_t gid, const FXbool rec, const FXbool dironly)
{
    struct stat stats;

	// Initialise the file variable with the initial path
	strlcpy(file, path, strlen(path)+1);

    if (lstatout(path, &stats)) //if it doesn't exist
        return -1;

    if (!S_ISDIR(stats.st_mode)) // file
    {
        if(dironly)
            return 0;

		// Force timeout checking for progress dialog
		checkTimeout();

		// Give cancel button an opportunity to be clicked
		if(cancelButton)
			getApp()->runModalWhileEvents(cancelButton);

		// Set labels for progress dialog
		FXString label=_("Changing owner...");
		uplabel->setText(label);
		label=_("File : ")+FXString(path);
		downlabel->setText(label);
		getApp()->repaint();

		// If cancel button was clicked, return
		if (isCancelled)
			return -1;

        if(::chown(path,uid,gid))
            return -1;
    }
    else // directory
    {
        if(rec == FALSE)
        {
			// Force timeout checking for progress dialog
			checkTimeout();

			// Give cancel button an opportunity to be clicked
			if(cancelButton)
				getApp()->runModalWhileEvents(cancelButton);

			// Set labels for progress dialog
			FXString label=_("Changing owner...");
			uplabel->setText(label);
			label=_("Folder : ")+FXString(path);
			downlabel->setText(label);
			getApp()->repaint();

			// If cancel button was clicked, return
			if (isCancelled)
				return -1;

            if(::chown(path,uid,gid)) // do not change recursively
                return -1;
        }
        else
            if(rchown(path,file,uid,gid,dironly)) // recursive change
                return -1;
    }
    return 0;

}


// Recursive chown for a directory
int File::rchown(char *path, char *file, uid_t uid, gid_t gid, const FXbool dironly)
{
    struct stat stats;

	// Initialise the file variable with the initial path
	strlcpy(file, path, strlen(path)+1);

    if (lstatout(path, &stats)) //if it doesn't exist
        return -1;

    if (!S_ISDIR(stats.st_mode)) // file
    {
		if(dironly)
			return 0;

		// Force timeout checking for progress dialog
		checkTimeout();

		// Give cancel button an opportunity to be clicked
		if(cancelButton)
			getApp()->runModalWhileEvents(cancelButton);

		// Set labels for progress dialog
		FXString label=_("Changing owner...");
		uplabel->setText(label);
		label=_("File : ")+FXString(path);
		downlabel->setText(label);
		getApp()->repaint();

		// If cancel button was clicked, return
		if (isCancelled)
			return -1;

		return ::chown(path,uid,gid);
    }

    DIR *dir;
    struct dirent *entry;
    int i, pl = strlen(path);

    if (!(dir = opendir(path)))
        return -1;

    for(i = 0; (entry = readdir(dir)); i++)
        if (entry->d_name[0] != '.' || (entry->d_name[1] != '\0'
                                        && (entry->d_name[1] != '.' ||
                                            entry->d_name[2] != '\0')))
        {
            int pl1 = pl, l = strlen(entry->d_name);
            char *path1 = (char *)alloca(pl1+l+2);

            strlcpy(path1, path, strlen(path)+1);
            if (path1[pl1-1] != '/')
                path1[pl1++] = '/';
            strlcpy(path1+pl1, entry->d_name, strlen(entry->d_name)+1);
			strlcpy(file, path1, strlen(path1)+1);
            if (rchown(path1,file,uid,gid,dironly))
                return -1;
        }

    if (closedir(dir))
        return -1;

    if(::chown(path,uid,gid))
        return -1;

    return 0;
}


// Extract an archive in a specified directory
int File::extract(const FXString name, const FXString dir, const FXString cmd)
{
	// Change to the specified directory
    chdir(dir.text());

    // Make and show command window
	CommandWindow *cmdwin=new CommandWindow(getApp(),_("Extract archive"),cmd,30,80);			
	cmdwin->create();

	// The command window object deletes itself after closing the window!

    return 1;
}


// Create an archive
int File::archive(const FXString name, const FXString cmd)
{
	// Target file already exists
	if(::exists(name))
    {
		// Overwrite dialog if necessary
		if (!overwrite_all & !skip_all)
		{
			FXString msg=_("File ")+name+_(" already exists. Overwrite?");
       		ConfirmOverWriteBox* dlg = new ConfirmOverWriteBox(mainWindow,_("Confirm Overwrite"),msg);
       		FXuint answer=dlg->execute();
			delete dlg;
       		switch(answer)
       		{
       			case 1:
           			overwrite = TRUE;
           			break;
       			case 2:
           			overwrite_all = TRUE;
           			break;
       			case 3:
           			overwrite = FALSE;
					break;
       			case 4:
           			skip_all = TRUE;
					break;
       		}
		}
		if(!(overwrite | overwrite_all) | skip_all)
			return FALSE;
	}

    // Make and show command window
	CommandWindow *cmdwin=new CommandWindow(getApp(),_("Add to archive"),cmd,30,80);			
	cmdwin->create();

	// The command window object deletes itself after closing the window!

    return 1;
}


#if defined(linux)
int File::mount(const FXString dir, const FXString msg, const FXString cmd, const unsigned int op)
{
	// Show progress dialog (no timer here)
	show(PLACEMENT_OWNER);
	getApp()->forceRefresh();
	getApp()->flush();

	// Set labels for progress dialog
	uplabel->setText(msg);				
	downlabel->setText(dir.text());
	getApp()->repaint();

	// Give cancel button an opportunity to be clicked
	if(cancelButton)
		getApp()->runModalWhileEvents(cancelButton);

	// If cancel button was clicked, return
	if (isCancelled)
		return -1;

	// Perform the mount/unmount command
	FILE *pcmd=popen(cmd.text(),"r");
	if(!pcmd)
	{
		MessageBox::error(this,MBOX_OK,_("Error"),_("Failed command : %s"),cmd.text());
		return -1;
	}

	// Get error message if any
	char text[10000]={0};
	FXString buf;
	while(fgets(text,sizeof(text),pcmd))
		buf+=text;
	snprintf(text,sizeof(text)-1,"%s",buf.text());
	
	// Close the stream
	if(pclose(pcmd))
	{
		MessageBox::error(this,MBOX_OK,_("Error"),"%s",text);
		return -1;
	}
	// Hide progress dialog
	hide();
	
	// Success message, eventually
	if(mount_messages)
	{
		if (op==MOUNT)
			MessageBox::information(this,MBOX_OK,_("Success"),_("Folder %s was successfully mounted."),dir.text());
		else
			MessageBox::information(this,MBOX_OK,_("Success"),_("Folder %s was successfully unmounted."),dir.text());
	}
	return 1;
}


// Install / Upgrade RPM package
int File::rpmInstall(const FXString name, const FXString cmd)
{
    // Make and show command window
	CommandWindow *cmdwin=new CommandWindow(getApp(),_("Install/Upgrade RPM"),cmd,10,80);			
	cmdwin->create();

	// The command window object deletes itself after closing the window!

    return 1;
}

// Uninstall RPM package
int File::rpmUninstall(const FXString name, const FXString cmd)
{
    // Make and show command window
	CommandWindow *cmdwin=new CommandWindow(getApp(),_("Uninstall RPM"),cmd,10,80);			
	cmdwin->create();
	
	// Add a message if necessary
	if (cmdwin->getLength()==0)
	{
		FXString msg=_("Package ") + name + _(" uninstalled successfully."); 
		cmdwin->appendText(msg.text());
	}

	// The command window object deletes itself after closing the window!

    return 1;
}

#endif


// Handle cancel button in progress bar dialog
long File::onCmdCancel(FXObject*, FXSelector,void*)
{
	isCancelled=TRUE;
	return 1;
}


// Handle timeout for progress bar
long File::onTimeout(FXObject*, FXSelector,void*)
{
	if (allowPdialog)
		show(PLACEMENT_OWNER);
    getApp()->forceRefresh();
    getApp()->flush();
    return 1;
}

